#!/usr/bin/env python
import os
import xarray
import pyproj
import numpy
import progressbar

from pyremap import ProjectionGridDescriptor, LatLonGridDescriptor, Remapper

"""
A script for combining the BedMachineAntarctica data set with GEBCO_2022.

Download the BedMachine topography
----------------------------------
1. Go to https://nsidc.org/data/nsidc-0756/versions/3
2. Click on "HTTPS File System" under "Data Access & Tools"
3. If you don't yet have an account, choose "Register"
4. Under "1970.01.01" choose "BedMachineAntarctica-v3.nc"

Download GEBCO
--------------
1. Go to https://www.gebco.net/data_and_products/gridded_bathymetry_data/
2. Click on "GEBCO_2022 Grid (sub-ice topo/bathy)"
3. unzip the zip file once downloaded


To run this script, using a compass conda and spack environment is best but
you can also create a conda environment as follows:
mamba create -y -n bedmachine python=3.10 pyproj pyremap numpy xarray netcdf4 \
    dask progressbar2 "esmf=*=*nompi*"

On Chrysalis, the following worked for me
srun -N 1 -t 1:00:00 --pty bash
source ~/mpas-work/compass/main/load_dev_compass_1.2.0-alpha.4_chrysalis_intel_openmpi.sh
srun ./combine_bedmachine_v3_gebco_2022.py
"""

esmf_mpi_tasks = 128

# range of lat to blend over
latmin = -62.
latmax = -60.

# the threshold for masks below which interpolated variables are not
# renormalized
renorm_thresh = 1e-3

# average GEBCO to 0.05 degree grid
# GEBCO is on 15" grid, so average every 12x12 block of cells

gebco_filename = 'GEBCO_2022_0.0125_degree.nc'
if os.path.exists(gebco_filename):
    gebco_low = xarray.open_dataset(gebco_filename)
else:
    gebco = xarray.open_dataset('GEBCO_2022_sub_ice_topo.nc')

    nlon = gebco.sizes['lon']
    nlat = gebco.sizes['lat']


    block = 3
    norm = 1.0/block**2

    nx = nlon//block
    ny = nlat//block

    chunks = 2
    nxchunk = nlon//chunks
    nychunk = nlat//chunks

    gebco = gebco.chunk({'lon': nxchunk, 'lat': nychunk})

    bathymetry = numpy.zeros((ny, nx))

    print('Averaging GEBCO to 0.0125 degree grid')
    widgets = [progressbar.Percentage(), ' ',
               progressbar.Bar(), ' ', progressbar.ETA()]
    bar = progressbar.ProgressBar(widgets=widgets,
                                  maxval=chunks**2).start()
    for ychunk in range(chunks):
        for xchunk in range(chunks):
            gebco_chunk = gebco.isel(
                lon=slice(nxchunk*xchunk, nxchunk*(xchunk+1)),
                lat=slice(nychunk*ychunk, nychunk*(ychunk+1)))
            elevation = gebco_chunk.elevation.values
            nxblock = nxchunk//block
            nyblock = nychunk//block
            bathy_block = numpy.zeros((nyblock, nxblock))
            for y in range(block):
                for x in range(block):
                    bathy_block += elevation[y::block, x::block]
            bathy_block *= norm
            bathymetry[ychunk*nyblock:(ychunk+1)*nyblock,
                       xchunk*nxblock:(xchunk+1)*nxblock] = bathy_block
            bar.update(ychunk*chunks+xchunk+1)
    bar.finish()

    lon_corner = numpy.linspace(-180., 180., bathymetry.shape[1]+1)
    lat_corner = numpy.linspace(-90., 90., bathymetry.shape[0]+1)
    lon = 0.5*(lon_corner[0:-1] + lon_corner[1:])
    lat = 0.5*(lat_corner[0:-1] + lat_corner[1:])
    gebco_low = xarray.Dataset({'bathymetry': (['lat', 'lon'], bathymetry)},
                               coords={'lon': (['lon',], lon),
                                       'lat': (['lat',], lat)})
    gebco_low.attrs = gebco.attrs
    gebco_low.lon.attrs = gebco.lon.attrs
    gebco_low.lat.attrs = gebco.lat.attrs
    gebco_low.bathymetry.attrs = gebco.elevation.attrs
    gebco_low.to_netcdf(gebco_filename)

# Modify BedMachineAntarctica to compute the fields needed by MPAS-Ocean

bedmachine_mod_filename = 'BedMachineAntarctica-v3_mod.nc'
if os.path.exists(bedmachine_mod_filename):
    bedmachine = xarray.open_dataset(bedmachine_mod_filename)
else:
    bedmachine = xarray.open_dataset('BedMachineAntarctica-v3.nc')
    mask = bedmachine.mask
    ice_mask = (mask != 0).astype(float)
    ocean_mask = (numpy.logical_or(mask == 0, mask == 3)).astype(float)
    grounded_mask = numpy.logical_or(numpy.logical_or(mask == 1, mask == 2),
                                     mask == 4).astype(float)

    bedmachine['bathymetry'] = bedmachine.bed.where(ocean_mask, 0.)
    bedmachine['ice_draft'] = \
        (bedmachine.surface - bedmachine.thickness).where(ocean_mask, 0.)
    bedmachine.ice_draft.attrs['units'] = 'meters'
    bedmachine['thickness'] = bedmachine.thickness.where(ocean_mask, 0.)

    bedmachine['ice_mask'] = ice_mask
    bedmachine['grounded_mask'] = grounded_mask
    bedmachine['ocean_mask'] = ocean_mask

    varlist = ['bathymetry', 'ice_draft', 'thickness', 'ice_mask',
               'grounded_mask', 'ocean_mask']

    bedmachine = bedmachine[varlist]

    bedmachine.to_netcdf(bedmachine_mod_filename)

projection = pyproj.Proj('+proj=stere +lat_ts=-71.0 +lat_0=-90 +lon_0=0.0 '
                         '+k_0=1.0 +x_0=0.0 +y_0=0.0 +ellps=WGS84')

lat_lon_projection = pyproj.Proj(proj='latlong', datum='WGS84')

bedmachine1k_filename = 'BedMachineAntarctica-v3_1k.nc'
if os.path.exists(bedmachine1k_filename):
    bedmachine1k = xarray.open_dataset(bedmachine1k_filename)
else:
    x = bedmachine.x.values
    y = bedmachine.y.values

    nx = len(x)//2
    ny = len(y)//2
    x = 0.5*(x[0:2*nx:2] + x[1:2*nx:2])
    y = 0.5*(y[0:2*ny:2] + y[1:2*ny:2])
    bedmachine1k = xarray.Dataset()
    for field in bedmachine.data_vars:
        inArray = bedmachine[field].values
        outArray = numpy.zeros((ny, nx))
        for yoffset in range(2):
            for xoffset in range(2):
               outArray += 0.25*inArray[yoffset:2*ny:2, xoffset:2*nx:2]
        da = xarray.DataArray(outArray, dims=('y', 'x'), 
                              coords={'x': (('x',), x),
                                      'y': (('y',), y)})
        bedmachine1k[field] = da
        bedmachine1k[field].attrs = bedmachine[field].attrs
      
    bedmachine1k.to_netcdf(bedmachine1k_filename)

x = bedmachine1k.x.values
y = bedmachine1k.y.values

out_filename = 'BedMachineAntarctica_on_GEBCO_low.nc'
if os.path.exists(out_filename):
    bedmachine_on_gebco_low = xarray.open_dataset(out_filename)
else:
    inDescriptor = ProjectionGridDescriptor.create(projection, x, y,
                                                   'BedMachineAntarctica1km')

    outDescriptor = LatLonGridDescriptor.read(ds=gebco_low)

    mappingFileName = 'map_BedMachineAntarctica1km_to_GEBCO_0.0125deg_bilinear.nc'

    remapper = Remapper(inDescriptor, outDescriptor, mappingFileName)
    remapper.build_mapping_file(method='bilinear', mpiTasks=esmf_mpi_tasks,
                                esmf_parallel_exec='srun', tempdir='.')
    bedmachine_on_gebco_low = remapper.remap(bedmachine1k)

    for field in ['bathymetry', 'ice_draft', 'thickness']:
        bedmachine_on_gebco_low[field].attrs['unit'] = 'meters'

    # renormalize the fields based on the ocean masks
    ocean_mask = bedmachine_on_gebco_low.ocean_mask

    valid = ocean_mask > renorm_thresh
    norm = ocean_mask.where(valid, 1.)
    norm = 1./norm
    for field in ['bathymetry', 'ice_draft', 'thickness']:
        bedmachine_on_gebco_low[field] = \
            norm*bedmachine_on_gebco_low[field].where(valid, 0.)

    bedmachine_on_gebco_low.to_netcdf(out_filename)

combined = xarray.Dataset()
alpha = (gebco_low.lat - latmin)/(latmax - latmin)
alpha = numpy.maximum(numpy.minimum(alpha, 1.0), 0.0)

bedmachine_bathy = bedmachine_on_gebco_low.bathymetry
valid = bedmachine_bathy.notnull()
bedmachine_bathy = bedmachine_bathy.where(valid, 0.)

combined['bathymetry'] = \
    alpha*gebco_low.bathymetry.where(gebco_low.bathymetry < 0, 0.) + \
    (1.0 - alpha)*bedmachine_bathy
for field in ['ice_draft', 'thickness']:
    combined[field] = bedmachine_on_gebco_low[field]
for field in ['bathymetry', 'ice_draft', 'thickness']:
    combined[field].attrs['unit'] = 'meters'

fill = {'ice_mask': 0., 'grounded_mask': 0.,
        'ocean_mask': combined['bathymetry'] < 0.}

for field in fill.keys():
    valid = bedmachine_on_gebco_low[field].notnull()
    combined[field] = bedmachine_on_gebco_low[field].where(valid, fill[field])

combined['water_column'] = combined['ice_draft'] - combined['bathymetry']
combined.water_column.attrs['units'] = 'meters'

combined.to_netcdf('BedMachineAntarctica_v3_and_GEBCO_2022_0.0125_degree_20230315.nc')

diff = xarray.Dataset()

diff['bathymetry'] = \
    gebco_low.bathymetry.where(gebco_low.bathymetry < 0, 0.) - \
    bedmachine_on_gebco_low.bathymetry
diff.to_netcdf('gebco_minus_bedmachine.nc')

